/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

#include "math.h"
#include <vector>
using std::vector;
using std::size_t;

#include "qubx_mac_dataprep.h"
#include "qub_adaptive_resample.h"
#include "qubx_idealize_mdl.h"
#include "qub_mean.h"

using std::max;
 
// congruent e.g.
//   signals: [current, ligand1, ligand2, voltage]
//   defaults: [    0,   1,       1,       -50] // only used for i>len(signals)
//   combineDelta: [0,  .1,      .1]            // only used for 0<i<len(signals)

extern "C" QUBFAST_API void qubx_mac_data_append_chunk_signals(void *data, int seg_ix, double dur_sec, double t0, double dt,
							       int Nsignal, int Nd, float **signals,
							       int Nstimulus, float *defaults, float *combineDelta)
{
  // storage for stimulus idealizations ([0] is for current because 0 means no stimulus in the rate matrices)
  vector<int> dwellCounts(Nstimulus);
  vector< vector<int> > classeses(Nstimulus);
  vector< vector<float> > durationses(Nstimulus);
  vector< vector<double> > ampses(Nstimulus);

  // set up time and current
  vector<float> T(Nd); // assuming no adaptive resample yet
  for (int i=0; i<Nd; ++i) {
    T[i] = (float) (i*dt);
  }
  float *I = signals[0];

  // use MDL to idealize sampled stimulus signals
  void *idlContext = qubx_idealize_mdl_context_create();
  vector<int> counts;
  vector<float> amps;
  for (int i=1; i<Nsignal; ++i) {
    int i1 = i-1;
    void *idl = qubx_idealizer_mdl_create(idlContext);
    int Ndwell = qubx_idealizer_mdl_idealize(idl, signals[i], Nd, 1, 0, 1, combineDelta[i]);
    dwellCounts[i1] = Ndwell;
    classeses[i1].resize(Ndwell);
    durationses[i1].resize(Ndwell);
    ampses[i1].resize(Ndwell);
    if ( (int) counts.size() < Ndwell ) {
      counts.resize(Ndwell);
      amps.resize(Ndwell);
    }
    qubx_idealizer_mdl_get_dwells(idl, & counts[0], &amps[0], 0, Ndwell);
    qubx_idealizer_mdl_free(idl);

    for (int j=0; j<Ndwell; ++j) {
      durationses[i1][j] = (float) (1e3 * dt * counts[j]);
      ampses[i1][j] = amps[j];
      classeses[i1][j] = j;
    }
  }
  qubx_idealize_mdl_context_free(idlContext);

  // other stimuli: generate constant idealizations from defaults
  for (int i=Nsignal; i<=Nstimulus; ++i) {
    classeses[i-1].push_back(0);
    durationses[i-1].push_back((float) (1e3 * dur_sec));
    ampses[i-1].push_back(defaults[i]);
  }

  vector<int*> classesesP(Nstimulus);
  vector<float*> durationsesP(Nstimulus);
  vector<double*> ampsesP(Nstimulus);
  for (int i=0; i<Nstimulus; ++i) {
    classesesP[i] = & classeses[i][0];
    durationsesP[i] = & durationses[i][0];
    ampsesP[i] = & ampses[i][0];
  }
  qubx_mac_data_append_chunk(data, seg_ix, dur_sec, t0, Nd,
			     & T[0], I, & dwellCounts[0],
			     & classesesP[0], & durationsesP[0], & ampsesP[0]);
}


class IdlClassifier {
public:
  double delta, deltaRSS;
  vector<IncrementalRSS> dist;

  IdlClassifier(double _delta)
    : delta(_delta), deltaRSS(_delta*_delta), dist(1)
  {}

  // durations in same units as sampling
  int add(double sampling, int Ndwell, float *durations, double *amps, double *stds, int *classes) {
    int Ncls = ((int) dist.size()) - 1;
    for (int i=0; i<Ndwell; ++i) {
      IncrementalRSS& event = dist[Ncls];
      int n = (int) (0.5 + durations[i]/sampling);
      event.set(amps[i], n*stds[i]*stds[i], n);
      double limitRSS = max(event.rss/n, deltaRSS);
      int c = -1;
      for (int j=0; j<Ncls; ++j) {
	if ( fabs(event.mean - dist[j].mean) < delta ) {
	  IncrementalRSS together = event + dist[j];
	  if ( (together.rss/n < limitRSS) || (together.rss/n < dist[j].rss/n) ) {
	    dist[j] = together;
	    c = j;
	    break;
	  }
	}
      }
      if ( c < 0 ) {
	c = Ncls++;
	dist.resize(Ncls+1);
      }
      classes[i] = c;
    }
    return Ncls;
  }

  int read(int Nclass, double *amps, double *stds) {
    if ( Nclass > (((int)dist.size())-1) ) {
      Nclass = ((int) dist.size()) - 1;
    }
    for (int i=0; i<Nclass; ++i) {
      amps[i] = dist[i].mean;
      stds[i] = dist[i].std();
    }
    return Nclass;
  }
};

extern "C" QUBFAST_API void* qubx_mac_idlcls_create(double delta) {
  return new IdlClassifier(delta);
}

extern "C" QUBFAST_API void qubx_mac_idlcls_free(void *obj) {
  delete ((IdlClassifier *) obj);
}

extern "C" QUBFAST_API int qubx_mac_idlcls_add(void *obj, double sampling, int Ndwell, float *durations, double *amps, double *stds, int *classes) {
  return ((IdlClassifier *) obj)->add(sampling, Ndwell, durations, amps, stds, classes);
}

extern "C" QUBFAST_API int qubx_mac_idlcls_read(void *obj, int Nclass, double *amps, double *stds) {
  return ((IdlClassifier *) obj)->read(Nclass, amps, stds);
}

